# -*- coding:utf-8 -*-

'''
Created on 2015/1/2

@author: junchen
'''
import numpy as np

def irt_fnc(theta, beta, alpha=1.0, c=0.0):
    # beta is item difficulty
    # theta is respondent capability

    prob = c+ (1.0-c) / (1 + np.exp(-(alpha*theta+beta)))
    return prob




def log_likelihood_factor_gradient(y1, y0, theta, alpha, beta, c=0.0):
    temp = np.exp(beta + alpha * theta)
    grad = alpha/(1.0+temp) *( y1*( (1.0-c)/(c/temp+1.0))-
                                  y0*temp)

    return grad

def log_likelihood_factor_hessian(y1, y0, theta, alpha, beta, c=0.0):
    x = np.exp(beta + alpha * theta)
    #hessian = - alpha**2*(y1+y0)*temp/(1+temp)**2
    hessian  =  alpha**2*x/(1+x)**2*( y1*(1-c)*(c-x**2)/(c+x)**2 -y0)

    return hessian


def logsum(logp):
    w = max(logp)
    logSump = w+ np.log(sum(np.exp(logp-w)))
    return logSump





import unittest
import os, sys
root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, root_dir)

import solver

class TestBayesianEstimator(unittest.TestCase):
    bad_log = [[1,(1,8,0)]] # not very informative
    good_log= [[0,(1,8,0)]] # very informative


    def test_bayesian_informative_log(self):
        uni_est = solver.theta_estimator.bayesian_estimator()
        uni_est.set_prior(-4,4,9,'uniform')

        beta_est = solver.theta_estimator.bayesian_estimator()
        beta_est.set_prior(-4,4,9,'beta')
        uni_est.update(self.good_log)

        uni_theta_posterior = uni_est.get_estimator()
        self.assertTrue(abs(uni_theta_posterior+3.413)<1e-2)

        beta_est.update(self.good_log)
        beta_theta_posterior = beta_est.get_estimator()
        self.assertTrue(abs(beta_theta_posterior+2.14)<1e-2)




    def test_bayesian_uninformative_log(self):
        uni_est = solver.theta_estimator.bayesian_estimator()
        uni_est.set_prior(-4,4,9,'uniform')

        beta_est = solver.theta_estimator.bayesian_estimator()
        beta_est.set_prior(-4,4,9,'beta')

        uni_est.update(self.bad_log)
        uni_theta_posterior = uni_est.get_estimator()
        self.assertTrue(abs(uni_theta_posterior-0.0)<1e-1)

        beta_est.update(self.bad_log)
        beta_theta_posterior = beta_est.get_estimator()
        self.assertTrue(abs(beta_theta_posterior-0.0)<1e-1)

    def test_mle_log(self):
        mle_est = solver.theta_estimator.MLE_estimator()
        theta_hat_bad = mle_est.update(self.bad_log)
        theta_hat_good = mle_est.update(self.good_log)

        self.assertTrue(abs(theta_hat_bad-3.69)<1e-2)
        self.assertTrue(abs(theta_hat_good+4.0)<1e-2)


if __name__ == '__main__':
    unittest.main()



'''
The model is an implementation of EM algorithm of IRT


For reference, see:
Brad Hanson, IRT Parameter Estimation using the EM Algorithm, 2000

The current version only deals with unidimension theta

'''
import numpy as np
import os
import sys
import time


root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, root_dir)
import utl
import solver


# import cython
class IRT_MMLE_2PL(object):

    '''
    Three steps are exposed
    (1) load data
    (2) set parameter
    (3) solve
    '''
    def load_data(self, src, is_mount, user_name, tmp_dir='/tmp/pyirt/'):
        # three columns are uid, eid, atag
        if isinstance(src, file):
            # if the src is file handle
            uids, eids, atags = self._loadFromHandle(src)
        else:
            # if the src is list of tuples
            uids, eids, atags = self._loadFromTuples(src)
        # process it
        print('Data loading is complete.')

        self.data_ref = utl.loader.data_storage()
        self.data_ref.setup(uids,eids,atags)

    def load_param(self, theta_bnds, alpha_bnds, beta_bnds):
        # TODO: allow for a more flexible parameter setting
        # The config object has to be passed in because hdfs file system does
        # not load target file

        # load user item
        num_theta = 11
        self._init_user_param(theta_bnds[0], theta_bnds[1], num_theta)

        # load the solver
        boundary = {'alpha': alpha_bnds,
                    'beta':  beta_bnds}

        solver_type    = 'gradient'
        is_constrained = True
        max_iter       = 10
        tol            = 1e-3

        self._init_solver_param(is_constrained, boundary, solver_type, max_iter, tol)

    def load_guess_param(self, in_guess_param):
        if isinstance(in_guess_param, basestring):
            # all c are 0
            guess_param_dict = {}
            for eid in self.data_ref.eid_vec:
                guess_param_dict[eid] = {'c': 0.0}
        else:
            guess_param_dict = in_guess_param

        self.guess_param_dict = guess_param_dict

    def solve_EM(self):
        # create the inner parameters
        # currently item parameter requires no setup
        self._init_item_param()

        self.posterior_theta_distr = np.zeros((self.data_ref.num_user, self.num_theta))

        # TODO: enable the stopping condition
        num_iter      = 1
        self.ell_list = []
        avg_prob_t0   = 0

        while True:
            iter_start_time = time.time()
            # add in time block
            start_time = time.time()
            self._exp_step()
            print("--- E step: %f secs ---" % np.round((time.time()-start_time)))

            start_time = time.time()
            self._max_step()
            print("--- M step: %f secs ---" % np.round((time.time()-start_time)))

            self.__calc_theta()

            '''
            Exp
            '''
            # self.update_guess_param()

            # the goal is to maximize the "average" probability
            avg_prob = np.exp(self.__calc_data_likelihood()/self.data_ref.num_log)
            self.ell_list.append(avg_prob)
            print("--- all: %f secs ---" % np.round((time.time()-iter_start_time)))
            print(avg_prob)

            # if the algorithm improves, then ell > ell_t0
            if avg_prob_t0 > avg_prob:
                # TODO: needs to roll back if the likelihood decrease
                print('Likelihood descrease, stops at iteration %d.' % num_iter)
                break

            if avg_prob_t0 < avg_prob and avg_prob - avg_prob_t0 <= self.tol:
                print('EM converged at iteration %d.' % num_iter)
                break
            # update the stop condition
            avg_prob_t0 = avg_prob
            num_iter += 1


            if (num_iter > self.max_iter):
                print('EM does not converge within max iteration')
                break

    def get_item_param(self):
        # need to remap the inner id to the outer id
        return self.item_param_dict

    def get_user_param(self):
        user_param_dict = {}
        for i in xrange(self.data_ref.num_user):
            uid = self.data_ref.uid_vec[i]
            user_param_dict[uid] = self.theta_vec[i]

        return user_param_dict

    '''
    Main Routine
    '''
    def _exp_step(self):
        '''
        Basic Math:
        In the maximization step, need to use E_[j,k](Y=1),E_[j,k](Y=0)
        E(Y=1|param_j,theta_k) = sum_i(data_[i,j]*P(Y=1|param_j,theta_[i,k]))
        since data_[i,j] = 0/1, it is equivalent to sum all done right users

        E(Y=0|param_j,theta_k) = sum_i(
                                (1-data_[i,j]) *(1-P(Y=1|param_j,theta_[i,k])
                                    )
        By similar logic, it is equivalent to sum (1-p) for all done wrong users

        '''

        # (1) update the posterior distribution of theta
        self.__update_theta_distr()

        # (2) marginalize
        # because of the sparsity, the expected right and wrong may not sum up
        # to the total num of items!
        self.__get_expect_count()

    def _max_step(self):
        '''
        Basic Math
            log likelihood(param_j) = sum_k(log likelihood(param_j, theta_k))
        '''
        # [A] max for item parameter
        opt_worker = solver.optimizer.irt_2PL_Optimizer()
        # the boundary is universal
        # the boundary is set regardless of the constrained option because the
        # constrained search serves as backup for outlier cases
        opt_worker.set_bounds([self.beta_bnds, self.alpha_bnds])

        # theta value is universal
        opt_worker.set_theta(self.theta_prior_val)

        for eid in self.data_ref.eid_vec:
            # set the initial guess as a mixture of current value and a new
            # start to avoid trap in local maximum
            initial_guess_val = (self.item_param_dict[eid]['beta'],
                                 self.item_param_dict[eid]['alpha'])

            opt_worker.set_initial_guess(initial_guess_val)
            opt_worker.set_c(self.item_param_dict[eid]['c'])

            # assemble the expected data
            j = self.data_ref.eidx[eid]
            expected_right_count = self.item_expected_right_bytheta[:,j]
            expected_wrong_count = self.item_expected_wrong_bytheta[:,j]
            input_data = [expected_right_count, expected_wrong_count]
            opt_worker.load_res_data(input_data)

            # solve by L-BFGS-B
            if self.solver_type == 'gradient':
                est_param = opt_worker.solve_param_gradient(self.is_constrained)
            elif self.solver_type == 'linear':
                # if the alogrithm is nelder-mead and the optimization fails to
                # converge, use the constrained version
                try:
                    est_param = opt_worker.solve_param_linear(self.is_constrained)
                except Exception  as  e:
                    if str(e) == 'Optimizer fails to find solution. Try constrained search.':
                        est_param = opt_worker.solve_param_linear(True)
                    else:
                        raise e
            else:
                raise Exception('Unknown solver type')

            # update
            self.item_param_dict[eid]['beta'] = est_param[0]
            self.item_param_dict[eid]['alpha'] = est_param[1]

        #### [B] max for theta density
        # pi = r_k/(w_k+r_k)
        r_vec = np.sum(self.item_expected_right_bytheta,axis=1)
        w_vec = np.sum(self.item_expected_wrong_bytheta,axis=1)
        self.theta_density = np.divide(r_vec, r_vec+w_vec)

    '''
    Auxuliary function
    '''
    def _loadFromTuples(self, data):
        uids  = []
        eids  = []
        atags = []
        if len(data) == 0:
            raise Exception('Data is empty.')

        for log in data:
            uids.append(int(log[0]))
            eids.append(int(log[1]))
            atags.append(int(log[2]))

        return uids, eids, atags

    def _loadFromHandle(self, fp, sep=','):
        # Default format is comma separated files,
        # Only int is allowed within the environment
        uids  = []
        eids  = []
        atags = []

        for line in fp:
            if line == '':
                continue
            uidstr, eidstr, atagstr = line.strip().split(sep)
            uids.append(int(uidstr))
            eids.append(int(eidstr))
            atags.append(int(atagstr))
        return uids, eids, atags

    def _init_solver_param(self, is_constrained, boundary,
                           solver_type, max_iter, tol):
        # initialize bounds
        self.is_constrained  = is_constrained
        self.alpha_bnds      = boundary['alpha']
        self.beta_bnds       = boundary['beta']
        self.solver_type     = solver_type
        self.max_iter        = max_iter
        self.tol             = tol

        if solver_type == 'gradient' and not is_constrained:
            raise Exception('BFGS has to be constrained')

    def _init_item_param(self):
        self.item_param_dict = {}
        for eid in self.data_ref.eid_vec:
            # need to call the old eid
            c = self.guess_param_dict[eid]['c']

            self.item_param_dict[eid] = {'alpha': 1.0, 'beta': 0.0, 'c': c}

    def _init_user_param(self, theta_min, theta_max, num_theta):
        self.theta_prior_val = np.linspace(theta_min, theta_max, num=num_theta)
        self.num_theta       = len(self.theta_prior_val)
        if self.num_theta   != num_theta:
            raise Exception('Theta initialization failed')
        # store the prior density
        self.theta_density   = np.ones(num_theta)/num_theta


    def __update_theta_distr(self):

        # TODO: consider shrinkage for the prior update
        '''
        Basic Math. Notice that the distribution is user specific
            P_t(theta,data_i,param) = p(data_i,param|theta)*p_[t-1](theta)
            p_t(data_i,param) = sum(p_t(theta,data_i,param)) over theta
            p_t(theta|data_i,param) = P_t(theta,data_i,param)/p_t(data_i,param)
        '''

        # [A] calculate p(data,param|theta)
        # TODO: speed it up
        for i in xrange(self.data_ref.num_user):
            uid = self.data_ref.uid_vec[i]
            # find all the items
            log_list = self.data_ref.get_log(uid)
            # create eid list and atag list
            num_log  = len(log_list)
            # create temp likelihood vector for each possible value of theta
            likelihood_vec= np.zeros(self.num_theta)
            # calculate
            for k in range(self.num_theta):
                theta     = self.theta_prior_val[k]
                # calculate the likelihood
                ell       = 0.0
                for m in range(num_log):
                    eid   = log_list[m][0]
                    alpha = self.item_param_dict[eid]['alpha']
                    beta  = self.item_param_dict[eid]['beta']
                    c     = self.item_param_dict[eid]['c']
                    atag  = log_list[m][1]
                    ell   += utl.clib.log_likelihood_2PL(atag, 1.0-atag,
                                                          theta, alpha, beta, c)
                # now update the density
                likelihood_vec[k] = ell

            # ell  = p(param|x), full joint = logp(param|x)+log(x)
            log_joint_prob_vec = likelihood_vec + np.log(self.theta_density)

            # calculate the posterior
            # p(x|param) = exp(logp(param,x) - log(sum p(param,x)))
            marginal = utl.tools.logsum(log_joint_prob_vec)
            self.posterior_theta_distr[i,:] = np.exp(log_joint_prob_vec - marginal)

        # When the loop finish, check if the theta_density adds up to unity for each user
        check_user_distr_marginal = np.sum(self.posterior_theta_distr, axis=1)
        if any(abs(check_user_distr_marginal - 1.0) > 0.0001):
            raise Exception('The posterior distribution of user ability is not proper')

    def __get_expect_count(self):

        self.item_expected_right_bytheta = np.zeros((self.num_theta, self.data_ref.num_item))
        self.item_expected_wrong_bytheta = np.zeros((self.num_theta, self.data_ref.num_item))

        for j in range(self.data_ref.num_item):
            eid = self.data_ref.eid_vec[j]
            # get all the users that done it right
            # get all the users that done it wrong
            right_uid_vec, wrong_uid_vec = self.data_ref.get_rwmap(eid)
            # condition on the posterior ability, what is the expected count of
            # students get it right
            # TODO: for readability, should specify the rows and columns
            self.item_expected_right_bytheta[:,j] = np.sum(self.posterior_theta_distr[right_uid_vec,:], axis = 0)
            self.item_expected_wrong_bytheta[:,j] = np.sum(self.posterior_theta_distr[wrong_uid_vec,:], axis = 0)


    def __calc_data_likelihood(self):
        # calculate the likelihood for the data set

        ell = 0
        for i in range(self.data_ref.num_user):
            uid = self.data_ref.uid_vec[i]
            theta = self.theta_vec[i]
            # find all the eid
            logs = self.data_ref.get_log(uid)
            for log in logs:
                eid = log[0]
                atag = log[1]
                alpha = self.item_param_dict[eid]['alpha']
                beta = self.item_param_dict[eid]['beta']
                c = self.item_param_dict[eid]['c']

                ell += utl.clib.log_likelihood_2PL(atag, 1-atag,
                                                    theta, alpha, beta, c)
        return ell

    def __calc_theta(self):
        self.theta_vec = np.dot(self.posterior_theta_distr, self.theta_prior_val)


    '''
    Experimental
    def update_guess_param(self):
        # at the end of each repetition, try to update the distribution of c

        #C is only identified at the extreme right tail, which is not possible in the EM environment
        #Use the average performance of the worst ability

        raise Exception('Currently deprecated!')

        # find the user that are in the bottom 5%
        cut_threshold = np.percentile(self.theta_vec, 5)
        bottom_group = [i for i in range(self.data_ref.num_user) if self.theta_vec[i] <= cut_threshold]

        # now loop through all the items
        for eid in self.data_ref.eid_vec:

            if self.item_param_dict[eid]['update_c']:
                # find the user group
                user_group = self.eid2uid_dict[eid]
                guessers = set(user_group).intersection(bottom_group)
                num_guesser = len(guessers)
                if num_guesser>10:
                    # average them
                    rw_list = self.right_wrong_map[eid]
                    right_cnt = 0.0
                    for uid in guessers:
                        if uid in rw_list['right']:
                            right_cnt += 1
                    # update c
                    # cap at 0.5
                    self.item_param_dict[eid]['c'] = min(right_cnt/num_guesser, 0.5)
    '''




import unittest
import os, sys
root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, root_dir)

import utl
import math
import numpy as np

class TestIrtFunctions(unittest.TestCase):


    def test_irt_fnc(self):
        # make sure the shuffled sequence does not lose any elements
        prob = utl.tools.irt_fnc(0.0,0.0,1.0)
        self.assertEqual(prob, 0.5)
        # alpha should play no role
        prob = utl.tools.irt_fnc(0.0,0.0,2.0)
        self.assertEqual(prob, 0.5)
        # higher theta should have higher prob
        prob = utl.tools.irt_fnc(1.0,0.0,1.0)
        self.assertEqual(prob, 1.0/(1.0+math.exp(-1.0)))
        # cancel out by higher beta
        prob = utl.tools.irt_fnc(1.0,-1.0,1.0)
        self.assertEqual(prob, 0.5)
        # test for c as limit situation
        prob = utl.tools.irt_fnc(-99,0.0,1.0, 0.25)
        self.assertTrue(abs(prob-0.25)<1e-5)
        prob = utl.tools.irt_fnc(99,0.0,1.0, 0.25)
        self.assertTrue(abs(prob-1.0)<1e-5)


    def test_log_likelihood(self):
        # raise error
        #with self.assertRaisesRegexp(ValueError,'Slope/Alpha should not be zero or negative.'):
        #     utl.clib.log_likelihood_2PL(0.0, 1.0, 0.0,-1.0,0.0)

        # the default model, log likelihood is log(0.5)
        ll = utl.clib.log_likelihood_2PL(1.0, 0.0, 0.0,1.0,0.0)
        self.assertEqual(ll, math.log(0.5))
        ll = utl.clib.log_likelihood_2PL(0.0, 1.0, 0.0,1.0,0.0)
        self.assertEqual(ll, math.log(0.5))

        # check the different model
        ll = utl.clib.log_likelihood_2PL(1.0,0.0, 1.0,1.0,0.0)
        self.assertEqual(ll, math.log(1.0/(1.0+math.exp(-1.0))))

        ll = utl.clib.log_likelihood_2PL(0.0,1.0, 1.0,1.0,0.0)
        self.assertEqual(ll, math.log(1.0-1.0/(1.0+math.exp(-1.0))))

        # check a real value
        ll = utl.clib.log_likelihood_2PL(0.0,1.0,-1.1617696779178492,1.0,0.0)

        self.assertTrue(abs(ll + 0.27226272946920399)<0.0000000001)

        # check if it handles c correctly
        ll = utl.clib.log_likelihood_2PL(1.0, 0.0, 0.0,1.0,0.0, 0.25)
        self.assertEqual(ll, math.log(0.625))
        ll = utl.clib.log_likelihood_2PL(0.0, 1.0, 0.0,1.0,0.0,0.25)
        self.assertEqual(ll, math.log(0.375))


    def test_log_sum(self):
        # add up a list of small values
        log_prob = np.array([-135,-115,-125,-100])
        approx_sum = utl.tools.logsum(log_prob)
        exact_sum = 0
        for num in log_prob:
            exact_sum += math.exp(num)
        exact_sum = math.log(exact_sum)
        self.assertTrue(abs(approx_sum-exact_sum)<1e-10)

    def test_log_item_gradient(self):
        delta = 0.00001
        y1 = 1.0
        y0 = 2.0
        theta = -2.0
        alpha = 1.0
        beta = 0.0
        # simulate the gradient
        true_gradient_approx_beta = (utl.clib.log_likelihood_2PL(y1,y0,theta,alpha,beta+delta) - \
            utl.clib.log_likelihood_2PL(y1,y0,theta,alpha,beta))/delta
        true_gradient_approx_alpha =  (utl.clib.log_likelihood_2PL(y1,y0,theta,alpha+delta,beta) - \
            utl.clib.log_likelihood_2PL(y1,y0,theta,alpha,beta))/delta
        # calculate
        calc_gradient = utl.clib.log_likelihood_2PL_gradient(y1,y0,theta,alpha,beta)

        self.assertTrue(abs(calc_gradient[0] - true_gradient_approx_beta ) < 1e-4)
        self.assertTrue(abs(calc_gradient[1] - true_gradient_approx_alpha) < 1e-4)

        # simulate the gradient with c
        c = 0.25
        true_gradient_approx_beta = (utl.clib.log_likelihood_2PL(y1,y0,theta,alpha,beta+delta,c) - \
            utl.clib.log_likelihood_2PL(y1,y0,theta,alpha,beta,c))/delta
        true_gradient_approx_alpha =  (utl.clib.log_likelihood_2PL(y1,y0,theta,alpha+delta,beta,c) - \
            utl.clib.log_likelihood_2PL(y1,y0,theta,alpha,beta,c))/delta
        # calculate
        calc_gradient = utl.clib.log_likelihood_2PL_gradient(y1,y0,theta,alpha,beta,c)

        self.assertTrue(abs(calc_gradient[0] - true_gradient_approx_beta ) < 1e-4)
        self.assertTrue(abs(calc_gradient[1] - true_gradient_approx_alpha) < 1e-4)


    def test_log_factor_gradient(self):
        delta = 0.00001
        y1 = 1.0
        y0 = 2.0
        theta = -2.0
        alpha = 1.0
        beta = 0.0
        # simulate the gradient
        true_gradient_approx_theta = (utl.clib.log_likelihood_2PL(y1,y0,theta+delta,alpha,beta) - \
            utl.clib.log_likelihood_2PL(y1,y0,theta,alpha,beta))/delta
       # calculate
        calc_gradient = utl.tools.log_likelihood_factor_gradient(y1,y0,theta,alpha,beta)

        self.assertTrue(abs(calc_gradient - true_gradient_approx_theta ) < 1e-4)


        # simulate the gradient
        c = 0.25
        true_gradient_approx_theta = (utl.clib.log_likelihood_2PL(y1,y0,theta+delta,alpha,beta,c) - \
            utl.clib.log_likelihood_2PL(y1,y0,theta,alpha,beta,c))/delta
       # calculate
        calc_gradient = utl.tools.log_likelihood_factor_gradient(y1,y0,theta,alpha,beta,c)

        self.assertTrue(abs(calc_gradient - true_gradient_approx_theta ) < 1e-4)


    def test_log_factor_hessian(self):
        delta = 0.00001
        y1 = 1.0
        y0 = 2.0
        theta = -2.0
        alpha = 1.0
        beta = 0.0
        c = 0.25
        # simulate the gradient
        true_hessian_approx_theta = (utl.tools.log_likelihood_factor_gradient(y1,y0,theta+delta,alpha,beta,c) - \
            utl.tools.log_likelihood_factor_gradient(y1,y0,theta,alpha,beta,c))/delta
       # calculate
        calc_hessian = utl.tools.log_likelihood_factor_hessian(y1,y0,theta,alpha,beta,c)

        self.assertTrue(abs(calc_hessian - true_hessian_approx_theta ) < 1e-4)

if __name__ == '__main__':
    unittest.main()



import os, sys
root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.insert(0, root_dir)

from pyirt import *
#LAST7_data = utl.loader.from_matrix_to_list('data/LAST7.txt',sep='\t',is_uid=True)
src_handle = open('data/sim_data_simple.txt','r')
'''
# load file handle
print('Load file handle.')
item_param,user_param = irt(src_handle)
src_handle.close()

# load tuples
print('Load tuple data.')
src_data = []
for line in src_handle:
    if line == '':
        continue
    uidstr, eidstr, atagstr = line.strip().split(',')
    src_data.append((int(uidstr),int(eidstr),int(atagstr)))

item_param,user_param = irt(src_data)
'''
# mount the damn tmp folder into memory
print('Mount into memory')
item_param,user_param = irt(src_handle, is_mount=True, user_name="junchen")

utl.loader.parse_item_paramer(item_param, output_file = 'data/sim_est.txt')




import os
import sys

root_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, root_dir)

import solver

def irt(src, theta_bnds=[-4,4],
        alpha_bnds = [0.25,2], beta_bnds=[-2,2], in_guess_param = 'default',
        model_spec='2PL',
        is_mount = False, user_name = None):

    if model_spec == '2PL':
        model = solver.model.IRT_MMLE_2PL()
    else:
        raise Exception('Unknown model specification.')

    # load
    model.load_data(src, is_mount, user_name)
    model.load_param(theta_bnds, alpha_bnds, beta_bnds)
    model.load_guess_param(in_guess_param)

    # solve
    model.solve_EM()

    # post
    item_param_dict = model.get_item_param()
    user_param_dict = model.get_user_param()


    return item_param_dict, user_param_dict




import unittest
import os, sys
root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, root_dir)
import solver, utl
import math
import numpy as np



class TestItemSolverNoGuess(unittest.TestCase):
    def setUp(self):
        # initialize the data
        n = 10000
        self.alpha = 1.5
        self.beta = -2.0
        theta_vec=np.random.normal(loc = 0.0, scale =1.0, size = n)
        y1 = []
        y0 = []
        for i in range(n):
            # generate the two parameter likelihood
            prob = utl.tools.irt_fnc(theta_vec[i], self.beta, self.alpha)
            # generate the response sequence
            if prob >= np.random.uniform():
                y1.append(1.0)
                y0.append(0.0)
            else:
                y1.append(0.0)
                y0.append(1.0)
        # the response format follows the solver API
        response_data = [y1, y0]

        # initialize optimizer
        self.solver = solver.optimizer.irt_2PL_Optimizer()
        self.solver.load_res_data(response_data)
        self.solver.set_theta(theta_vec)
        self.solver.set_c(0)

    def test_linear_unconstrained(self):
        self.solver.set_initial_guess((0.0,1.0))
        est_param = self.solver.solve_param_linear(is_constrained = False)
        self.assertTrue(abs(est_param[0]-self.beta)<0.07 and abs(est_param[1]-self.alpha)<0.05)

    def test_linear_constrained(self):
        self.solver.set_initial_guess((0.0,1.0))
        self.solver.set_bounds([(-4.0,4.0),(0.25,2)])
        est_param = self.solver.solve_param_linear(is_constrained = True)
        self.assertTrue(abs(est_param[0]-self.beta)<0.05 and abs(est_param[1]-self.alpha)<0.05)



    #def test_gradient_unconstrained(self):
    #    self.solver.set_initial_guess((0.0,2.0))
    #    est_param = self.solver.solve_param_gradient(is_constrained = False)
    #    import ipdb; ipdb.set_trace()  # XXX BREAKPOINT

    def test_gradient_constrained(self):
        self.solver.set_initial_guess((0.0,1.0))
        self.solver.set_bounds([(-4.0,4.0),(0.25,2)])
        est_param = self.solver.solve_param_gradient(is_constrained = True)
        self.assertTrue(abs(est_param[0]-self.beta)<0.1 and abs(est_param[1]-self.alpha)<0.1)



class TestUserSolver(unittest.TestCase):
    def setUp(self):
        # initialize the data
        n = 1000
        self.theta = 1.5
        self.alpha_vec = np.random.random(n)+1
        self.beta_vec  = np.random.normal(loc = 0.0, scale =2.0, size = n)
        self.c_vec = np.zeros(n)
        y1 = []
        y0 = []
        for i in range(n):
            # generate the two parameter likelihood
            prob = utl.tools.irt_fnc(self.theta, self.beta_vec[i], self.alpha_vec[i])
            # generate the response sequence
            if prob >= np.random.uniform():
                y1.append(1.0)
                y0.append(0.0)
            else:
                y1.append(0.0)
                y0.append(1.0)
        # the response format follows the solver API
        response_data = [y1, y0]

        # initialize optimizer
        self.solver = solver.optimizer.irt_factor_optimizer()
        self.solver.load_res_data(response_data)
        self.solver.set_item_parameter(self.alpha_vec, self.beta_vec, self.c_vec)
        self.solver.set_bounds([(-6,6)])

    def test_linear_unconstrained(self):
        self.solver.set_initial_guess(0.0)
        est_param = self.solver.solve_param_linear(is_constrained = False)
        self.assertTrue(abs(est_param-self.theta)<0.1)

    def test_gradient_constrained(self):
        self.solver.set_initial_guess(0.0)
        self.solver.set_bounds([(-6,6)])
        est_param = self.solver.solve_param_gradient(is_constrained = True)
        self.assertTrue(abs(est_param-self.theta)<0.1)

    def test_hessian_unconstrained(self):
        self.solver.set_initial_guess(0.0)
        est_param = self.solver.solve_param_hessian()
        self.assertTrue(abs(est_param-self.theta)<0.1)


    def test_scalar_constrained(self):
        self.solver.set_initial_guess(0.0)
        self.solver.set_bounds((-6,6))
        est_param = self.solver.solve_param_scalar()
        self.assertTrue(abs(est_param-self.theta)<0.1)


if __name__ == '__main__':
    unittest.main()




import numpy as np
from scipy.stats import beta


import os, sys
root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, root_dir)

import utl, solver

# import cython
import pyximport; pyximport.install()
import utl.clib as clib

class bayesian_estimator(object):

    def set_prior(self, theta_min, theta_max, num_theta, dist_name):
        self.theta_val = np.linspace(theta_min, theta_max, num = num_theta)
        self.num_theta = num_theta
        if dist_name == 'uniform':
            # the density is uniform
            self.theta_density = np.ones(num_theta)/num_theta

        elif dist_name == 'beta':
            # centered beta
            # rescale to move away from the boundary
            self.theta_density = beta.pdf((self.theta_val-theta_min)/(theta_max-theta_min+0.1),2,2)
            # renormalize
            self.theta_density = self.theta_density /sum(self.theta_density )
        else:
            raise Exception('Unknown prior distribution.')


    def update(self, logs):
        # data comes in as
        # tag(0/1), (a, b,c)

        likelihood_vec = np.zeros(self.num_theta)
        # calculate
        for k in range(self.num_theta):
            theta = self.theta_val[k]
            # calculate the likelihood
            ell = 0.0
            for log in logs:
                atag  = log[0]
                alpha = log[1][0]
                beta  = log[1][1]
                c     = log[1][2]
                ell += clib.log_likelihood_2PL(atag, 1.0-atag, theta, alpha, beta, c)
            # now update the density
            likelihood_vec[k] = ell

        # ell  = p(param|x), full joint = logp(param|x)+log(x)
        log_joint_prob_vec  = likelihood_vec + np.log(self.theta_density)
        # calculate the posterior
        # p(x|param) = exp(logp(param,x) - log(sum p(param,x)))
        marginal = utl.tools.logsum(log_joint_prob_vec)
        self.theta_density = np.exp(log_joint_prob_vec - marginal)

    def get_estimator(self):

        # expected value
        theta_mean = np.dot(self.theta_density, self.theta_val)
        #theta_var = np.dot(self.theta_density, self.theta_val**2) - theta_mean**2

        theta_hat = theta_mean
        return theta_hat


class MLE_estimator(object):
    worker = solver.optimizer.irt_factor_optimizer()

    def update(self, logs):
        # log [tag(0/1), (a, b,c)]

        # transform the logs
        y1     = []
        y0     = []
        alphas = []
        betas  = []
        cs     = []
        for log in logs:
            y1.append(log[0])
            y0.append(1.0-log[0])
            alphas.append(log[1][0])
            betas.append(log[1][1])
            cs.append(log[1][2])
        self.worker.load_res_data([y1,y0])
        self.worker.set_item_parameter(alphas, betas, cs)
        self.worker.set_bounds([(-4.0,4.0)])
        self.worker.set_initial_guess(0.0)
        try:
            est_theta = self.worker.solve_param_gradient(is_constrained=True)
        except:
            est_theta = self.worker.solve_param_linear(is_constrained=True)
        # the output is an numpy array!
        return est_theta[0]




import numpy as np
from scipy.optimize import minimize
from scipy.optimize import minimize_scalar

import os, sys
root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, root_dir)
import utl



# TODO: The BFGS method is not as precise as the NM method
# TODO: There maybe overflowing issue in data
# TODO: Enable the calibration for two parameter estimation
np.seterr(over='raise')

class irt_2PL_Optimizer(object):

    def load_res_data(self, res_data):
        self.res_data = np.array(res_data)

    def set_theta(self, theta):
        self.theta = theta

    def set_c(self, c):
        self.c = c

    def set_bounds(self, bnds):
        self.bnds = bnds

    def set_initial_guess(self, x0):
        self.x0 = x0

    # generate the likelihood function
    @staticmethod
    def _likelihood(res_data, theta_vec, alpha, beta, c):
        # for MMLE method, y1 and y0 will be expected count
        y1 = res_data[0]
        y0 = res_data[1]

        # check for equal length between y1,y0 and theta_vec
        num_data = len(y1)
        if len(y0) != num_data:
            raise ValueError('The response data does not match in length.')
        if len(theta_vec) != num_data:
            raise ValueError('The response data does not match theta vec in length.')

        if sum(y1<0)>0 or  sum(y0<0)>0:
            raise ValueError('y1 or y0 contains negative count.')
        # this is the likelihood
        likelihood_vec = [utl.clib.log_likelihood_2PL(y1[i],y0[i],theta_vec[i],
                                                     alpha, beta,c) \
                          for i in range(num_data)]
        # transform into negative likelihood
        ell = -sum(likelihood_vec)

        return ell

    @staticmethod
    def _gradient(res_data, theta_vec, alpha, beta, c):
        # res should be numpy array
        y1 = res_data[0]
        y0 = res_data[1]
        num_data = len(y1)

        der = np.zeros(2)
        for i in range(num_data):
            # the toolbox calculate the gradient of the log likelihood,
            # but the algorithm needs that of the negative ll
            der -= utl.clib.log_likelihood_2PL_gradient(y1[i],y0[i],theta_vec[i],alpha,beta, c)
        #grad = [ -utl.clib.log_likelihood_2PL_gradient(y1[i],y0[i],theta_vec[i],alpha,beta) for i in range(num_data)]
        #TODO: This is actually a bit of cheating
        #if abs(der[0]) >50 or abs(der[1])>50:
        #    ratio = max(abs(der/50))
        #    der = der/ratio
        return der


    def solve_param_linear(self, is_constrained):
        # for now, temp set alpha to 1
        def target_fnc(x):
            beta = x[0]
            alpha = x[1]
            return self._likelihood(self.res_data, self.theta, alpha, beta, self.c)

        if is_constrained:
            res = minimize(target_fnc, self.x0, method = 'SLSQP',
                        bounds=self.bnds, options={'disp':False})
        else:
            res = minimize(target_fnc, self.x0, method='nelder-mead',
                           options={'xtol':1e-3, 'disp':False})

        # deal with expcetions
        if not res.success:
            if not is_constrained and \
                    res.message == 'Maximum number of function evaluations has been exceeded.':
                raise Exception('Optimizer fails to find solution. Try constrained search.')
            else:
                raise Exception('Algorithm failed because '+ res.message)

        return res.x

    def solve_param_gradient(self, is_constrained):
        # for now, temp set alpha to 1

        def target_fnc(x):
            beta = x[0]
            alpha = x[1]
            return self._likelihood(self.res_data, self.theta, alpha, beta, self.c)

        def target_der(x):
            beta = x[0]
            alpha = x[1]
            return self._gradient(self.res_data, self.theta, alpha, beta, self.c)

        if is_constrained:
            res = minimize(target_fnc, self.x0, method = 'L-BFGS-B',
                        jac= target_der, bounds = self.bnds,
                        options={'disp':False})
        else:
            res = minimize(target_fnc, self.x0, method = 'BFGS',
                           jac=target_der,
                           options={'disp':False})

        if not res.success:
            raise Exception('Algorithm failed.')

        return res.x

class irt_factor_optimizer(object):
    def load_res_data(self, res_data):
        self.res_data = np.array(res_data)

    def set_item_parameter(self, alpha_vec, beta_vec, c_vec):
        if len(alpha_vec) != len(beta_vec):
            raise ValueError('The alpha vec and the beta vec does not match in length.')

        self.alpha_vec = alpha_vec
        self.beta_vec  = beta_vec
        self.c_vec = c_vec

    def set_bounds(self, bnds):
        self.bnds = bnds

    def set_initial_guess(self, x0):
        self.x0 = x0

    @staticmethod
    def _likelihood(res_data, theta, alpha_vec, beta_vec, c_vec):
        # for MMLE method, y1 and y0 will be expected count
        y1 = res_data[0]
        y0 = res_data[1]

        # check for equal length between y1,y0 and theta_vec
        num_data = len(y1)
        if len(y0) != num_data:
            raise ValueError('The response data does not match in length.')
        if len(alpha_vec) != num_data:
            raise ValueError('The response data does not match alpha vec in length.')
        if sum(y1<0)>0 or  sum(y0<0)>0:
            raise ValueError('y1 or y0 contains negative count.')
        # this is the likelihood
        likelihood_vec = [utl.clib.log_likelihood_2PL(y1[i],y0[i],theta,
                                                     alpha_vec[i], beta_vec[i],c_vec[i]) \
                          for i in range(num_data)]
        # transform into negative likelihood
        ell = -sum(likelihood_vec)

        return ell

    @staticmethod
    def _gradient(res_data, theta, alpha_vec, beta_vec, c_vec):
        # res should be numpy array
        y1 = res_data[0]
        y0 = res_data[1]
        num_data = len(y1)

        der = 0.0
        for i in range(num_data):
            der -= utl.tools.log_likelihood_factor_gradient(y1[i],y0[i],theta,alpha_vec[i],beta_vec[i], c_vec[i])
        return der

    @staticmethod
    def _hessian(res_data, theta, alpha_vec, beta_vec, c_vec):
        # res should be numpy array
        y1 = res_data[0]
        y0 = res_data[1]
        num_data = len(y1)

        hes = 0.0
        for i in range(num_data):
            hes -= utl.tools.log_likelihood_factor_hessian(y1[i],y0[i],theta,alpha_vec[i],beta_vec[i], c_vec[i])
        return hes


    def solve_param_linear(self, is_constrained):
        # for now, temp set alpha to 1
        def target_fnc(x):
            return self._likelihood(self.res_data, x, self.alpha_vec, self.beta_vec, self.c_vec)

        if is_constrained:
            res = minimize(target_fnc, self.x0, method = 'SLSQP',
                        bounds=self.bnds, options={'disp':False})
        else:
            res = minimize(target_fnc, self.x0, method='nelder-mead',
                           options={'xtol':1e-4, 'disp':False})

        # deal with expcetions
        if not res.success:
            if not is_constrained and \
                    res.message == 'Maximum number of function evaluations has been exceeded.':
                raise Exception('Optimizer fails to find solution. Try constrained search.')
            else:
                raise Exception('Algorithm failed because '+ res.message)

        return res.x

    def solve_param_gradient(self, is_constrained):
        # for now, temp set alpha to 1
        def target_fnc(x):
            return self._likelihood(self.res_data, x, self.alpha_vec, self.beta_vec, self.c_vec)
        def target_der(x):
            return self._gradient(self.res_data, x, self.alpha_vec, self.beta_vec, self.c_vec)

        if is_constrained:
            res = minimize(target_fnc, self.x0, method = 'L-BFGS-B',
                        jac= target_der, bounds = self.bnds,
                        options={'disp':False})
        else:
            res = minimize(target_fnc, self.x0, method = 'BFGS',
                           jac=target_der,
                           options={'disp':False})

        if not res.success:
            raise Exception('Algorithm failed.')

        return res.x

    def solve_param_hessian(self):
        def target_fnc(x):
            return self._likelihood(self.res_data, x, self.alpha_vec, self.beta_vec, self.c_vec)
        def target_der(x):
            return self._gradient(self.res_data, x, self.alpha_vec, self.beta_vec, self.c_vec)
        def target_hess(x):
            return self._hessian(self.res_data, x, self.alpha_vec, self.beta_vec, self.c_vec)

        res = minimize(target_fnc, self.x0, method='Newton-CG',
                jac=target_der, hess=target_hess,
                options={'xtol': 1e-8, 'disp': False})
        if not res.success:
            if res.message == 'Desired error not necessarily achieved due to precision loss.':
                # TODO:still returns a result. Something is wrong with the BFGS
                # though
                pass
            else:
                raise Exception('Algorithm failed, because ' + res.message)
        return res.x


    def solve_param_scalar(self):
        def target_fnc(x):
            return self._likelihood(self.res_data, x, self.alpha_vec, self.beta_vec, self.c_vec)
        res = minimize_scalar(target_fnc, bounds=self.bnds, method='bounded')
        return res.x




import cython
import numpy as np

def log_likelihood_2PL(double y1,
                       double y0,
                       double theta,
                       double alpha,
                       double beta,
                       double c=0.0):
    cdef extern from "math.h":
        double exp(double x)
        double log(double x)
    expComp = exp(-(alpha*theta + beta));
    ell =  y1*log(c+(1.0-c)/(1.0+expComp)) + y0*log((1.0-c)*expComp/(1.0+expComp)) ;

    return ell


def log_likelihood_2PL_gradient(double y1,
                                double y0,
                                double theta,
                                double alpha,
                                double beta,
                                double c=0.0):
    #TODO: could be organized into matrix
    # It is the gradient of the log likelihood, not the NEGATIVE log likelihood
    cdef extern from "math.h":
        double exp(double x)

    grad = np.zeros(2)

    temp = exp(beta + alpha * theta)
    beta_grad = 1.0/(1.0+temp) *( y1*( (1.0-c)/(c/temp+1.0))-
                                  y0*temp)

    #beta_grad = -(-y1+y0*temp)/(1+temp)
    alpha_grad = theta*beta_grad
    grad[0] = beta_grad
    grad[1] = alpha_grad
    return grad



from setuptools import setup

setup(
  name = 'pyirt',
  version = '0.2.4',
  packages = ['pyirt',
		'pyirt/solver',
		'pyirt/utl',
		'pyirt/test',],
  url = 'https://github.com/junchenfeng/pyirt',
  license='MIT',
  description = 'A python implementation of EM IRT, specializing in sparse massive data',
  author = 'Junchen Feng',
  author_email = 'frankfeng.pku@gmail.com',
  include_package_data=True,
  download_url = 'https://github.com/junchenfeng/pyirt/tarball/0.2.4',
  keywords = ['IRT', 'EM', 'big data'],
  zip_safe=False,
  platforms='any',
  install_requires=[
   'numpy',
   'scipy',
   'cython',],
  package_data={'pyirt': ["*.pyx"]},
  classifiers = [
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
  ],
)



'''
This script deals with the data format problem.

The stardard format for pyirt is ( uid,eid,result),
where uid is the idx for test taker, eid is the idx for items

It is set in this way to deal with the sparsity in the massive dataset.

'''
import numpy as np
import time
import os
import subprocess
try:
    import bsddb
except:
    import bsddb3 as bsddb

import collections as cos

def from_matrix_to_list(indata_file, sep=',',header=False, is_uid=False):
    # assume the data takes the following format
    # (uid,) item1, item2, item3
    # (1,)   0,     1,     1
    is_skip = True
    is_init = False
    uid = 0

    result_list = []

    with open(indata_file,'r') as f:
        for line in f:
            if is_skip and header:
                # if there is a header, skip
                is_skip=False
                continue
            segs = line.strip().split(sep)

            if len(segs)==0:
                continue

            if not is_init:
                # calibrate item id
                if is_uid:
                    num_item = len(segs)-1
                else:
                    num_item = len(segs)

            # parse
            for j in range(num_item):
                if is_uid:
                    idx = j+1
                else:
                    idx = j
                result_list.append((uid, j, int(segs[idx])))

            # TODO: the current code is hard wired to uid starts from 0 to n
            # needs to remove the dependency
            uid += 1

    return result_list


def load_sim_data(sim_data_file):
    # this function loads the simulation data for testing
    # the sim format is
    #(result, uid, eid, theta, beta, alpha)
    test_data = []
    test_param = {}
    # the outputs are [a] solver readable dataset, [b] item parameter
    with open(sim_data_file, 'r') as f:
        for line in f:
            if line.strip() == '':
                continue

            result, uid, eid, theta, beta, alpha= line.strip().split(',')
            test_data.append((int(uid), int(eid), int(result)))
            if eid not in test_param:
                test_param[int(eid)]={'alpha':float(alpha), 'beta':float(beta)}
    return test_data, test_param

def load_dbm(dmb_val):
    # the format is 'id,flag;'*n
    pairs = dmb_val.split(';')
    log_list = []
    # last element is empty
    for i in xrange(len(pairs)-1):
        idstr,flagstr = pairs[i].split(',')
        log_list.append((int(idstr), int(flagstr)))
    return log_list




def parse_item_paramer(item_param_dict, output_file = None):

    if output_file is not None:
        # open the file
        out_fh = open(output_file,'w')

    sorted_eids = sorted(item_param_dict.keys())

    for eid in sorted_eids:
        param = item_param_dict[eid]
        alpha_val = np.round(param['alpha'],decimals=2)
        beta_val = np.round(param['beta'],decimals=2)
        if output_file is None:
            print eid, alpha_val, beta_val
        else:
            out_fh.write('{},{},{}\n'.format(eid, alpha_val, beta_val))

'''
Build a data storage facility that allows for memory dict and bsddb dict
'''

class data_storage(object):

    def setup(self, uids, eids, atags, mode='memory',
              tmp_dir = None, is_mount = False, user_name = None):
        # mode could be 'memory', which uses RAM, or 'dbm', which uses the hard disk.
        # When in 'dbm', set is_mount = False and the user_name to allows for
        # RAMdisk. However, it is not nearly as fast as memory when data is
        # large, unless uses SSD
        self.mode = mode

        if mode == 'dbm':
            # check if the tmp directory is accessible
            if not os.path.isdir(tmp_dir):
                os.mkdir(tmp_dir)
            # by default,do NOT mount the temp dir in memory, unless otherwise specified
            if is_mount:
                # create the directory
                subprocess.call(["sudo", "mount","-t","tmpfs","tmpfs",tmp_dir])
                # transfer ownwership
                subprocess.call(["sudo","chown",user_name+":root",tmp_dir])

            self.tmp_dir = tmp_dir  # passed in for later cache

        # pre processing
        start_time = time.time()
        if mode == 'memory':
            self._process_data_memory(uids, eids, atags)
        elif mode == 'dbm':
            self._process_data_dbm(uids, eids, atags)
        else:
            raise Exception('Unknown mode of data storage.')

        self._init_data_param()
        print("--- Process: %f secs ---" % np.round((time.time()-start_time)))


       # initialize some intermediate variables used in the E step
        start_time = time.time()
        if mode == 'memory':
            self._init_right_wrong_map_memory()
        elif mode == 'dbm':
            self._init_right_wrong_map_dbm()
        else:
            raise Exception('Unknown mode of data storage.')
        print("--- Sparse Mapping: %f secs ---" % np.round((time.time()-start_time)))

    '''
    Need the following dictionary for esitmation routine
    (1) item -> user: key: eid, value: (uid, atag)
    (2) user -> item: key: uid, value: (eid, atag)
    '''

    def _process_data_memory(self, uids, eids, atags):
        self.num_log = len(uids)
        self.item2user = cos.defaultdict(list)
        self.user2item = cos.defaultdict(list)

        for i in xrange(self.num_log):
            eid  = eids[i]
            uid  = uids[i]
            atag = atags[i]
            # add to the data dictionary
            self.item2user[eid].append((uid, atag))
            self.user2item[uid].append((eid, atag))

    def _process_data_dbm(self, uids, eids, atags):
        '''
        Memory efficiency optimization:

        (1) The matrix is sparse, so use three parallel lists for data storage.
        parallel lists are more memory eficient than tuple
        (2) For fast retrieval, turn three parallel list into dictionary by uid and eid
        (3) Python dictionary takes up a lot of memory, so use dbm

        # for more details, see
        http://stackoverflow.com/questions/2211965/python-memory-usage-loading-large-dictionaries-in-memory

        '''

        # always rewrite
        self.item2user = bsddb.hashopen(self.tmp_dir+'/item2user.db', 'n')
        self.user2item = bsddb.hashopen(self.tmp_dir+'/user2item.db', 'n')
        self.num_log = len(uids)

        for i in xrange(self.num_log):
            eid  = eids[i]
            uid  = uids[i]
            atag = atags[i]
            # if not initiated, init with empty str
            if str(eid) not in self.item2user:
                self.item2user['%d' % eid] = ''
            if str(uid) not in self.user2item:
                self.user2item['%d' % uid] = ''

            self.item2user['%d' % eid] += '%d,%d;' % (uid, atag)
            self.user2item['%d' % uid] += '%d,%d;' % (eid, atag)

    def _init_right_wrong_map_bdm(self):
        self.right_map = bsddb.hashopen(self.tmp_dir+'/right_map.db', 'n')
        self.wrong_map = bsddb.hashopen(self.tmp_dir+'/wrong_map.db', 'n')

        for eidstr, log_val_list in self.item2user.iteritems():
            log_result = utl.loader.load_dbm(log_val_list)
            for log in log_result:
                # The E step uses the index of the uid
                uid = log[0]
                atag = log[1]
                uid_idx = self.uidx[uid]
                if atag == 1:
                    if eidstr not in self.right_map:
                        self.right_map[eidstr] = '%d' % uid_idx
                    else:
                        self.right_map[eidstr] += ',%d' % uid_idx
                else:
                    if eidstr not in self.wrong_map:
                        self.wrong_map[eidstr] = '%d' % uid_idx
                    else:
                        self.wrong_map[eidstr] += ',%d' % uid_idx

    def _init_right_wrong_map_memory(self):
        self.right_map = cos.defaultdict(list)
        self.wrong_map = cos.defaultdict(list)

        for eid, log_result in self.item2user.iteritems():
            for log in log_result:
                atag = log[1]
                uid  = log[0]
                uid_idx = self.uidx[uid]
                if atag == 1:
                    self.right_map[eid].append(uid_idx)
                else:
                    self.wrong_map[eid].append(uid_idx)

    def _init_data_param(self):
        # system parameter
        self.uid_vec = [int(x) for x in self.user2item.keys()]
        self.num_user = len(self.uid_vec)
        self.eid_vec = [int(x) for x in self.item2user.keys()]
        self.num_item = len(self.eid_vec)

        # build a dictionary for fast uid index, which is used in map
        self.uidx = cos.defaultdict(int)
        for i in xrange(self.num_user):
            self.uidx[self.uid_vec[i]] = i

        self.eidx = cos.defaultdict(int)
        for j in xrange(self.num_item):
            self.eidx[self.eid_vec[j]] = j


    def get_log(self, uid):
        if self.mode == 'bdm':
            log_list = load_dbm(self.user2item_db['%d' % uid])
        elif self.mode == 'memory':
            log_list = self.user2item[uid]
        else:
            raise Exception('Unknown mode of storage.')
        return log_list

    def get_rwmap(self, eid):
        if self.mode == 'bdm':
            right_uid_vec = [int(x) for x in self.right_map[str(eid)].split(',') ]
            wrong_uid_vec = [int(x) for x in self.wrong_map[str(eid)].split(',') ]
        elif self.mode == 'memory':
            right_uid_vec = self.right_map[eid]
            wrong_uid_vec = self.wrong_map[eid]
        else:
            raise Exception('Unknown mode of storage.')
        return right_uid_vec, wrong_uid_vec



